import pytest

from pytest_mock_resources.compat import pymongo
from pytest_mock_resources.container import (
    ContainerCheckFailed,
    get_container_fn,
    get_docker_host,
    IN_CI,
)

# XXX: To become overwritable via pytest config.
config = {
    "port": 27017 if IN_CI else 28017,
    "root_database": "dev-mongo",
    "image": "mongo:3.6",
}


def get_pymongo_client():
    uri = "mongodb://{}:{}".format(get_docker_host(), config["port"])

    return pymongo.MongoClient(uri)


def check_mongo_fn():
    try:
        client = get_pymongo_client()
        db = client[config["root_database"]]
        db.command("ismaster")
    except pymongo.errors.ConnectionFailure:
        raise ContainerCheckFailed(
            "Unable to connect to a presumed MongoDB test container via given config: {}".format(
                config
            )
        )


_mongo_container = pytest.fixture("session")(
    get_container_fn(
        "_mongo_container", config["image"], {27017: config["port"]}, {}, check_mongo_fn
    )
)



import os
import socket
import time

import docker
import responses

from pytest_mock_resources.compat import functools

IN_CI = os.getenv("CI") == "true"  # type: bool


@functools.lru_cache()
def get_docker_host():
    host = "host.docker.internal"
    try:
        socket.gethostbyname(host)
        return host
    except socket.gaierror:
        return os.environ.get("PYTEST_MOCK_RESOURCES_HOST", "localhost")


class ContainerCheckFailed(Exception):
    """Unable to connect to a Container.
    """


def get_container_fn(name, image, ports, environment, check_fn):
    def wrapped():
        # XXX: moto library may over-mock responses. SEE: https://github.com/spulec/moto/issues/1026
        responses.add_passthru("http+docker")

        def retriable_check_fn(retries):
            while retries:
                retries -= 1
                try:
                    check_fn()
                    return
                except Exception:
                    if not retries:
                        raise
                    time.sleep(1)

        try:
            container = None
            try:
                retriable_check_fn(1)
            except ContainerCheckFailed:
                client = docker.from_env(version="auto")
                container = client.containers.run(
                    image, ports=ports, environment=environment, detach=True, remove=True
                )
                retriable_check_fn(20)

            yield
        except Exception:
            raise

        finally:
            if container:
                container.kill()

    wrapped.__name__ = name

    return wrapped


from pytest_mock_resources.container.mongo import _mongo_container  # noqa, isort:skip
from pytest_mock_resources.container.postgres import _postgres_container  # noqa, isort:skip



import pytest

from pytest_mock_resources.compat import pymongo
from pytest_mock_resources.container import get_docker_host
from pytest_mock_resources.container.mongo import config, get_pymongo_client
from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials


def create_mongo_fixture(**kwargs):
    scope = kwargs.pop("scope", "function")

    if len(kwargs):
        raise KeyError("Unsupported Arguments: {}".format(kwargs))

    @pytest.fixture(scope=scope)
    def _(_mongo_container):
        return _create_clean_database()

    return _


def _create_clean_database():
    client = get_pymongo_client()
    db = client[config["root_database"]]

    # Create a collection called `pytestMockResourceDbs' in the admin tab if it hasnt already been
    # created.
    db_collection = db["pytestMockResourcesDbs"]

    #  create a Document in the `pytestMockResourcesDbs` collection:
    result = db_collection.insert_one({})
    db_id = str(result.inserted_id)

    #  Create a database where the name is equal to that ID:
    create_db = client[db_id]

    #  Create a user as that databases owner
    create_db.command("createUser", db_id, pwd="password", roles=["dbOwner"])  # nosec

    #  pass back an authenticated db connection
    limited_client = pymongo.MongoClient(get_docker_host(), config["port"])
    db = limited_client[db_id]
    db.authenticate(db_id, "password")

    db.config = {"username": db_id, "password": "password", "database": db_id}

    assign_fixture_credentials(
        db,
        drivername="mongodb",
        host=get_docker_host(),
        port=config["port"],
        database=db_id,
        username=db_id,
        password="password",
    )

    return db



import pytest
import sqlalchemy

from pytest_mock_resources.container import (
    ContainerCheckFailed,
    get_container_fn,
    get_docker_host,
    IN_CI,
)

# XXX: To become overwritable via pytest config.
config = {
    "username": "user",
    "password": "password",
    "port": 5432 if IN_CI else 5532,
    "root_database": "dev",
    "image": "postgres:9.6.10-alpine",
}


def get_sqlalchemy_engine(database_name, isolation_level=None):
    URI_TEMPLATE = (
        "postgresql+psycopg2://{username}:{password}@{host}:{port}/{database}?sslmode=disable"
    )
    DB_URI = URI_TEMPLATE.format(
        database=database_name,
        username=config["username"],
        password=config["password"],
        host=get_docker_host(),
        port=config["port"],
    )

    options = {}
    if isolation_level:
        options["isolation_level"] = isolation_level

    # Trigger any psycopg2-based import failures
    from pytest_mock_resources.compat import psycopg2

    psycopg2.connect

    engine = sqlalchemy.create_engine(DB_URI, **options)

    # Verify engine is connected
    engine.connect()

    return engine


def check_postgres_fn():
    try:
        get_sqlalchemy_engine(config["root_database"])
    except sqlalchemy.exc.OperationalError:
        raise ContainerCheckFailed(
            "Unable to connect to a presumed Postgres test container via given config: {}".format(
                config
            )
        )


_postgres_container = pytest.fixture("session")(
    get_container_fn(
        "_postgres_container",
        config["image"],
        {5432: config["port"]},
        {
            "POSTGRES_DB": config["root_database"],
            "POSTGRES_USER": config["username"],
            "POSTGRES_PASSWORD": config["password"],
        },
        check_postgres_fn,
    )
)



import pytest
import sqlalchemy

from pytest_mock_resources.container import get_docker_host
from pytest_mock_resources.container.postgres import config, get_sqlalchemy_engine
from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials
from pytest_mock_resources.fixture.database.relational.generic import EngineManager


def create_postgres_fixture(*ordered_actions, **kwargs):
    """Create a Postgres fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.
    """
    scope = kwargs.pop("scope", "function")
    tables = kwargs.pop("tables", None)

    session = kwargs.pop("session", None)

    if len(kwargs):
        raise KeyError("Unsupported Arguments: {}".format(kwargs))

    @pytest.fixture(scope=scope)
    def _(_postgres_container):
        database_name = _create_clean_database()
        engine = get_sqlalchemy_engine(database_name)

        assign_fixture_credentials(
            engine,
            drivername="postgresql+psycopg2",
            host=get_docker_host(),
            port=config["port"],
            database=database_name,
            username="user",
            password="password",
        )

        engine_manager = EngineManager(
            engine, ordered_actions, tables=tables, default_schema="public"
        )
        for engine in engine_manager.manage(session=session):
            yield engine

    return _


def _create_clean_database():
    root_engine = get_sqlalchemy_engine(config["root_database"], isolation_level="AUTOCOMMIT")

    try:
        root_engine.execute(
            """
            CREATE TABLE IF NOT EXISTS pytest_mock_resource_db(
                id serial
            );
            """
        )
    except (sqlalchemy.exc.IntegrityError, sqlalchemy.exc.ProgrammingError):
        # A race condition may occur during table creation if:
        #  - another process has already created the table
        #  - the current process begins creating the table
        #  - the other process commits the table creation
        #  - the current process tries to commit the table creation
        pass

    result = root_engine.execute(
        "INSERT INTO pytest_mock_resource_db VALUES (DEFAULT) RETURNING id"
    )
    id_ = tuple(result)[0][0]
    database_name = "pytest_mock_resource_db_{}".format(id_)

    root_engine.execute('CREATE DATABASE "{}"'.format(database_name))
    root_engine.execute(
        'GRANT ALL PRIVILEGES ON DATABASE "{}" TO CURRENT_USER'.format(database_name)
    )

    return database_name



import pytest

from pytest_mock_resources.container import get_docker_host
from pytest_mock_resources.container.postgres import config
from pytest_mock_resources.fixture.database.generic import assign_fixture_credentials
from pytest_mock_resources.fixture.database.relational.generic import EngineManager
from pytest_mock_resources.fixture.database.relational.postgresql import (
    _create_clean_database,
    get_sqlalchemy_engine,
)
from pytest_mock_resources.patch.redshift.create_engine import (
    substitute_execute_with_custom_execute,
)


def create_redshift_fixture(*ordered_actions, **kwargs):
    """Create a Redshift fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.
    """
    scope = kwargs.pop("scope", "function")
    tables = kwargs.pop("tables", None)

    session = kwargs.pop("session", None)

    if len(kwargs):
        raise KeyError("Unsupported Arguments: {}".format(kwargs))

    from pytest_mock_resources.fixture.database.relational.redshift.udf import REDSHIFT_UDFS

    ordered_actions = ordered_actions + (REDSHIFT_UDFS,)

    @pytest.fixture(scope=scope)
    def _(_postgres_container):
        database_name = _create_clean_database()
        engine = get_sqlalchemy_engine(database_name)

        assign_fixture_credentials(
            engine,
            drivername="postgresql+psycopg2",
            host=get_docker_host(),
            port=config["port"],
            database=database_name,
            username="user",
            password="password",
        )

        engine = substitute_execute_with_custom_execute(engine)
        engine_manager = EngineManager(
            engine, ordered_actions, tables=tables, default_schema="public"
        )
        for engine in engine_manager.manage(session=session):
            yield engine

    return _



# flake8: noqa
from pytest_mock_resources.fixture.database.mongo import create_mongo_fixture
from pytest_mock_resources.fixture.database.relational import (
    create_postgres_fixture,
    create_redshift_fixture,
    create_sqlite_fixture,
    Rows,
    Statements,
)



# flake8: noqa
from pytest_mock_resources.fixture.database.relational.generic import Rows, Statements
from pytest_mock_resources.fixture.database.relational.postgresql import create_postgres_fixture
from pytest_mock_resources.fixture.database.relational.redshift import create_redshift_fixture
from pytest_mock_resources.fixture.database.relational.sqlite import create_sqlite_fixture



from sqlalchemy import create_engine
from sqlalchemy.sql.elements import TextClause

from pytest_mock_resources.compat import functools, mock, psycopg2
from pytest_mock_resources.container.postgres import config
from pytest_mock_resources.patch.redshift.create_engine import (
    execute_mock_s3_copy_command,
    execute_mock_s3_unload_command,
)
from pytest_mock_resources.patch.redshift.mock_s3_copy import strip


class CustomCursor(psycopg2.extensions.cursor):
    """A custom cursor class to define a custom execute method."""

    def execute(self, sql, args=None):
        dsn_params = self.connection.get_dsn_parameters()
        engine = create_engine(
            "postgresql+psycopg2://{user}:{password}@{hostname}:{port}/{dbname}".format(
                user=dsn_params["user"],
                password=config["password"],
                hostname=dsn_params["host"],
                port=dsn_params["port"],
                dbname=dsn_params["dbname"],
            )
        )
        if not isinstance(sql, TextClause) and strip(sql).lower().startswith("copy"):
            return execute_mock_s3_copy_command(sql, engine)
        if not isinstance(sql, TextClause) and strip(sql).lower().startswith("unload"):
            return execute_mock_s3_unload_command(sql, engine)
        return super(CustomCursor, self).execute(sql, args)


def patch_psycopg2_connect(path):
    """Patch any occourances of `psycopg2.connect` with mock_psycopg2_connect function.

    The `path` should be the path to the `psycopg2` module you want to patch.
    """

    class MockPsycoPg2:
        """Avoid an infinite circular mock scenario.

        Mocks the whole of psycopg2, so as to not accidentally infinitely mock the mocked `connect`.
        """

        def __getattr(self, name):
            return getattr(psycopg2, name)

        def connect(self, *args, **kwargs):
            return mock_psycopg2_connect(*args, **kwargs)

    def _patch_psycopg2_connect(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with mock.patch(path, new=MockPsycoPg2()):
                return func(*args, **kwargs)

        return wrapper

    return _patch_psycopg2_connect


def mock_psycopg2_connect(*args, **kwargs):
    """Substitute the default cursor with a custom cursor."""
    conn = psycopg2.connect(*args, cursor_factory=CustomCursor, **kwargs)
    return conn



from sqlalchemy.engine.url import URL


class Credentials:
    def __init__(self, drivername, host, port, database, username, password):
        self.drivername = drivername
        self.host = host
        self.port = port
        self.database = database
        self.username = username
        self.password = password

    def as_url(self):
        return str(self.as_url())

    def as_sqlalchemy_url(self):
        return URL(
            drivername=self.drivername,
            host=self.host,
            port=self.port,
            database=self.database,
            username=self.username,
            password=self.password,
        )

    def as_psycopg2_kwargs(self):
        return {
            "host": self.host,
            "port": self.port,
            "user": self.username,
            "password": self.password,
            "dbname": self.database,
        }

    def as_mongo_kwargs(self):
        return {
            "host": self.host,
            "port": self.port,
            "username": self.username,
            "password": self.password,
            "authSource": self.database,
        }


def assign_fixture_credentials(engine, **credentials):
    engine.pmr_credentials = Credentials(**credentials)



# flake8: noqa
from pytest_mock_resources.container import _mongo_container, _postgres_container  # noqa
from pytest_mock_resources.fixture.database import (
    create_mongo_fixture,
    create_postgres_fixture,
    create_redshift_fixture,
    create_sqlite_fixture,
    Rows,
    Statements,
)
from pytest_mock_resources.patch import patch_create_engine, patch_psycopg2_connect
